import cmath
import math
import random

precision = 12
_digits = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"}
openBrackets = {"(", "{", "["}
closeBrackets = {")", "}", "]"}
funcTwoArgs = {"mod", "rand", "randint"}
funcMultArgs = funcTwoArgs.union({"sum", "mean", "product", "min", "max", "range",
                        "median", "mode", "variance", "stdev", "gcd", "lcm", "pow"})
funcComplexArgs = {"sin", "cos", "tan", "asin", "acos", "atan", "csc", "sec", "cot", 
"ln", "log", "abs", "sqr", "exp", "sinh", "cosh", "tanh", "asinh", "acosh", "atanh",
"ceil", "floor", "trunc", "sum", "mean", "product", "mode", "variance", "stdev", "gcd",
                                                        "lcm", "mod", "pow", "fac"}
functions = {"sin", "cos", "tan", "asin", "acos", "atan", "csc", "sec", "cot", 
"ln", "log", "abs", "sqr", "exp", "sinh", "cosh", "tanh", "asinh", "acosh", "atanh",
"ceil", "floor", "trunc", "sum", "mean", "product", "min", "max", "range", "median",
"mode", "variance", "stdev", "gcd", "lcm", "mod", "pow", "rand", "randint", "fac"}

# Function Definitions

def gcd(array):
        if len(array) == 0:
                return 1
        elif len(array) == 1:
                return array[0]
        elif len(array) == 2:
                a = array[0]
                b = array[1]
                if isinstance(a, complex) or isinstance(b, complex):
                        return gcd([complex(str(a)).real, complex(str(a)).imag, complex(str(b)).real, complex(str(b)).imag])
                a = abs(a)
                b = abs(b)
                if a == 0 or b == 0:
                        return max(a, b)
                if (a % b == 0 or b % a == 0):
                        return min(a, b)
                newNum = max(a, b) - (long)(max(a, b) / min(a, b)) * min(a, b)
                return gcd([newNum, min(a, b)])
        else:
                return gcd([gcd(array[:len(array) / 2]), gcd(array[len(array) / 2:])])

def lcm(array):
        if len(array) == 0:
                return 0
        if len(array) == 1:
                return array[0]
        elif len(array) == 2:
                return array[0] * array[1] / gcd(array)
        else:
                return lcm([lcm(array[:len(array) / 2]), lcm(array[len(array) / 2:])])

# Coefficients used by the GNU Scientific Library
_g = 7
_p = [0.99999999999980993, 676.5203681218851, -1259.1392167224028,
     771.32342877765313, -176.61502916214059, 12.507343278686905,
     -0.13857109526572012, 9.9843695780195716e-6, 1.5056327351493116e-7]

def gamma(z):
        # Uses Lanczos Approximation to calculate Gamma Function
        z = complex(z)
        # Reflection formula
        if z.real < 0.5:
                return cmath.pi / (cmath.sin(cmath.pi * z) * gamma(1 - z))
        else:
                z -= 1
                x = _p[0]
                for i in range(1, _g + 2):
                        x += _p[i] / (z + i)
                t = z + _g + 0.5
        return cmath.sqrt(2 * cmath.pi) * (t ** (z + 0.5)) * cmath.exp(-t) * x

# Private Function Evaluation Helper Methods

def _replaceAltFuncSpellings(expression):
        expression = expression.replace("sine", "sin").replace("cosine", "cos").replace("tangent", "tan")
        expression = expression.replace("cosecant", "csc").replace("secant", "sec").replace("cotangent", "cot")
        expression = expression.replace("arc", "a").replace("sqrt", "sqr").replace("integer", "int")
        expression = expression.replace("flr", "floor").replace("ceiling", "ceil").replace("truncate", "trunc")
        expression = expression.replace("average", "avg").replace("minimum", "min").replace("maximum", "max")
        expression = expression.replace("standarddev", "stdev").replace("power", "pow").replace("random", "rand")
        expression = expression.replace("modulo", "mod").replace("avg", "mean").replace("factorial", "fact")
        expression = expression.replace("fact", "fac")
        return expression

# Public Function Evaluation Methods

def oneArgFuncEval(function, value):
        # Evaluates functions that take a complex number input
        if function == "sin":
                return cmath.sin(value)
        elif function == "cos":
                return cmath.cos(value)
        elif function == "tan":
                return cmath.tan(value)
        elif function == "asin":
                return cmath.asin(value)
        elif function == "acos":
                return cmath.acos(value)
        elif function == "atan":
                return cmath.atan(value)
        elif function == "csc":
                return 1.0 / cmath.sin(value)
        elif function == "sec":
                return 1.0 / cmath.cos(value)
        elif function == "cot":
                return 1.0 / cmath.tan(value)        
        elif function == "ln":
                return cmath.log(value)
        elif function == "sqr":
                return cmath.sqrt(value)
        elif function == "abs":
                return cmath.sqrt((value.real ** 2) + (value.imag ** 2))
        elif function == "exp":
                return cmath.exp(value)
        if function == "sinh":
                return cmath.sinh(value)
        elif function == "cosh":
                return cmath.cosh(value)
        elif function == "tanh":
                return cmath.tanh(value)
        elif function == "asinh":
                return cmath.asinh(value)
        elif function == "acosh":
                return cmath.acosh(value)
        elif function == "atanh":
                return cmath.atanh(value)
        elif function == "ceil":
                return math.ceil(value.real) + complex(0, 1) * math.ceil(value.imag)
        elif function == "floor":
                return math.floor(value.real) + complex(0, 1) * math.floor(value.imag)
        elif function == "trunc":
                return math.trunc(value.real) + complex(0, 1) * math.trunc(value.imag)
        elif function == "fac":
                if value.imag == 0 and value < 0 and value.real == int(value.real):
                        return "Error: The factorial function is not defined on the negative integers."
                return gamma(value + 1)
        elif function == "log":
                return cmath.log10(value)
                                
def multArgsFuncEval(function, arguments):
        # Evaluates functions with multiple parameters
        if function == "sum" or function == "mean":
                answer = 0
                for arg in arguments:
                        answer += arg
                if function == "sum":
                        return answer
                return answer / len(arguments)
        elif function == "product":
                answer = 1
                for arg in arguments:
                        answer *= arg
                return answer
        elif function == "min":
                return min(arguments)
        elif function == "max":
                return max(arguments)
        elif function == "range":
                return max(arguments) - min(arguments)
        elif function == "median":
                return sorted(arguments)[len(arguments) / 2]
        elif function == "mode":
                return max(set(arguments), key = arguments.count)
        elif function == "variance" or function == "stdev":
                sum = 0
                for arg in arguments:
                        sum += arg
                mean = sum / len(arguments)
                answer = 0
                for arg in arguments:
                        answer += (arg - mean) ** 2
                if function == "variance":
                        return answer / len(arguments)
                return cmath.sqrt(answer / len(arguments))
        elif function == "gcd":
                return gcd(arguments)
        elif function == "lcm":
                return lcm(arguments)
        elif function == "mod":
                firstNum = arguments[0]
                secondNum = arguments[1]
                return firstNum + (math.ceil((-firstNum / secondNum).real) + complex(0, 1) * math.ceil((-firstNum / secondNum).imag)) * secondNum
        elif function == "pow":
                answer = arguments[0]
                for i in xrange(1, len(arguments)):
                        answer **= arguments[i]
                return answer
        elif function == "log":
                if len(arguments) == 1:
                        return cmath.log10(arguements[0])
                elif len(arguments) == 2:
                        return cmath.log(arguments[0], arguements[1])
                else:
                        return "Error: 'log' has too many arguments."
        elif function == "rand":
                return random.uniform(arguments[0], arguments[1])
        elif function == "randint":
                return random.randint(arguments[0], arguments[1])
                
                        
def evalFunc(expression, function):
        for digit in _digits:
                expression = expression.replace(digit + function, digit + "*" + function)
        for bracket in closeBrackets:
                expression = expression.replace(bracket + function, bracket + "*" + function)
        expression = expression.replace("i" + function, "i" + "*" + function).replace("j" + function, "j" + "*" + function)
        
        # Finds the opening and closing brackets for the special mathematical function
        while expression.find(function) != -1:
                functionIndex = expression.find(function)
                bracketCount = 1
                openBracketIndex = functionIndex + len(function)
                closeBracketIndex = openBracketIndex + 1
                for char in expression[openBracketIndex + 1:]:
                        if char in openBrackets:
                                bracketCount += 1
                        elif char in closeBrackets:
                                bracketCount -= 1
                        if bracketCount == 0:
                                break
                        closeBracketIndex += 1
                answer = 0
                
                # Evaluates subexpressions
                for bracket in openBrackets:
                        if expression[openBracketIndex + 1 : closeBracketIndex].find(bracket) != -1:
                                beginning = expression[:openBracketIndex + 1]
                                end = expression[closeBracketIndex:]
                                subExpr = evaluateFunctions(expression[openBracketIndex + 1 : closeBracketIndex])
                                expression = beginning + subExpr + end
                                closeBracketIndex = len(beginning) + len(subExpr)

                if function in funcMultArgs:
                        # Separates the parameteres of any function with multiple arguments into a list and evaluates function
                        args = [x.strip() for x in expression[openBracketIndex + 1 : closeBracketIndex].split(',')]
                        if len(args) == 0 or (len(args) < 2 and function in funcTwoArgs):
                                return "Error: '" + function + "' is missing arguments."
                        elif function not in funcMultArgs and len(args) != 1:
                                return "Error: '" + function + "' only takes one argument."
                        elif len(args) > 2 and function in funcTwoArgs:
                                return "Error: '" + function + "' has too many arguments."
                        for i in xrange(0, len(args)):
                                args[i] = complex(amp.evaluate(str(args[i])).replace("i", "j"))
                                if function not in funcComplexArgs:
                                        if complex(args[i]).imag != 0:
                                                return "Error: '" + function + "' does not except nonreal arguments."
                                        args[i] = args[i].real
                        answer = multArgsFuncEval(function, args)
                else:
                        # Evaluates the expression inside the brackets
                        value = complex(amp.eval(expression[openBracketIndex + 1 : closeBracketIndex]).replace("i", "j"))
                        if function not in funcComplexArgs:
                                if value.imag != 0:
                                        return "Error: '" + function + "' does not except nonreal arguments."                                       
                                value = value.real
                        answer = oneArgFuncEval(function, value)

                expression = expression[:functionIndex] + amp.complexFormat(answer) + expression[closeBracketIndex + 1:]
        return expression

def evaluateFunctions(expression):
        expression = _replaceAltFuncSpellings(expression)
        for func in functions:
                expression = evalFunc(expression, func)
        return expression

import amp
amp.precision = precision + 5



#                    _       ____    ____  _______
#                   / \     |_   \  /   _||_   __ \
#                  / _ \      |   \/   |    | |__) |
#                 / ___ \     | |\  /| |    |  ___/ 
#               _/ /   \ \_  _| |_\/_| |_  _| |_ 
#              |____| |____||_____||_____||_____|
# 
#                Automatic Mathematical Parser
#                       By Ini Oguntola
#
# Parses and evaluates any numerical mathematical expression, 
# with support for complex numbers and mathematical constants, as well 
# as scientific notation and a wide range of mathematical functions.

import math
import cmath
import decimal
import funcEval

precision = 12
funcEval.precision = precision
_operatorSet = {"+", "-", "*", "/", "%", "^", "!"}
_digits = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"}
openBrackets = {"(", "{", "["}
closeBrackets = {")", "}", "]"}

# Private Implementation Helper Methods

def _indexOfSpecificCharacters(string, characters):
        return next((i for i, char in enumerate(string) if char in characters), -1)

def _indexOfSpecificOperators(expression, operators):
        # Skips the first character when searching for operators
        index = _indexOfSpecificCharacters(expression[1:], operators) + 1
        if index == 0:
                return -1
        return index

def _indexOfOperator(expression):
        return _indexOfSpecificOperators(expression, _operatorSet)
        
def _indexOfPreviousCharacters(expression, characters, currentIndex):
        index = -1
        for char in characters:
                if expression[:currentIndex].rfind(char) > index:
                        index = expression[:currentIndex].rfind(char)
        return index
        
def _indexOfPreviousOperator(expression, currentIndex):
        # Iterates until the next index is the index you're looking for, then returns the index of the operator before it
        lastIndex = -1
        nextIndex = _indexOfSpecificOperators(expression, _operatorSet)
        while nextIndex < currentIndex and nextIndex != lastIndex:
                lastIndex = nextIndex
                nextIndex = _indexOfSpecificOperators(expression[nextIndex + 1:], _operatorSet) + nextIndex + 1
                if nextIndex == lastIndex:
                        break
        return lastIndex

def _pemdas(expression, operators):
        while _indexOfSpecificOperators(expression, operators) != -1:
                # Returns the current expression if all that's left is a complex number
                if expression.find(">") > expression.rfind("<"):
                        if expression[0] == "<" and expression[len(expression) - 1] == ">":
                                return expression[1 : len(expression) - 1]
                        elif expression[:2] == "-<" and expression[len(expression) - 1] == ">":
                                return eval("-1*(" + expression[2 : len(expression) - 1] + ")")
                        
                # Finds the next operator of the given type, along with the operators before and after it
                if operators == {"^"}:
                        # Adjusted to account for exponents being right-associative
                        currentIndex = expression.rfind("^")
                        # Keeps precedence of power over unary negation by counting leading minus signs as operators
                        previousIndex = _indexOfPreviousCharacters(expression, _operatorSet, currentIndex)
                else:
                        currentIndex = _indexOfSpecificOperators(expression, operators)
                        previousIndex = _indexOfPreviousOperator(expression, currentIndex)
                currentOperator = expression[currentIndex : currentIndex + 1]
                nextIndex = _indexOfOperator(expression[currentIndex + 1:]) + currentIndex + 1
                if nextIndex == currentIndex:
                        nextIndex = len(expression)

                # Skips previous operator if operator is part of a complex number expression
                if previousIndex != -1 and previousIndex < _indexOfPreviousCharacters(expression, {">"}, currentIndex):
                        previousIndex = _indexOfPreviousCharacters(expression[:expression[:previousIndex].rfind("<")], _operatorSet, previousIndex)

                # Skips current operator if the operator is part of a complex number expression (the +/- in "a +/- bi")
                if expression.find(">") != -1 and _indexOfOperator(expression[expression.find(">"):]) > 0 and currentIndex > expression.find("<") and currentIndex < expression.find(">"):
                        currentIndex = _indexOfOperator(expression[expression.find(">"):]) + expression.find(">")
                        currentOperator = expression[currentIndex : currentIndex + 1]
                        nextIndex = _indexOfOperator(expression[currentIndex + 1:]) + currentIndex + 1
                        if nextIndex == currentIndex:
                                nextIndex = len(expression)
                                
                # Skips next operator if operator is part of a complex number expression
                if expression[currentIndex : nextIndex].find("<") != -1:
                        tempNext = nextIndex + expression[nextIndex:].find(">")
                        nextIndex += _indexOfOperator(expression[expression[nextIndex:].find(">") + nextIndex:]) + expression[nextIndex:].find(">")
                        if nextIndex == tempNext - 1:
                                nextIndex = len(expression)
                             
                # Replaces all "i"s with "j"s to be parsed as complex numbers
                expression = expression[:previousIndex + 1] + expression[previousIndex + 1 : nextIndex].replace("i", "j") + expression[nextIndex:] 

                # Parses the string to find values in the expression at the operator
                firstNum = complex(expression[previousIndex + 1 : currentIndex].replace("<", "").replace(">", "").replace("--", ""))
                secondNum = 0
                try:
                        secondNum = complex(expression[currentIndex + 1 : nextIndex].replace("<", "").replace(">", "").replace("--", ""))
                except:
                        pass
                result = 0

                # Calculates result based on operator and values and substitutes it into the expression
                if currentOperator == "^":
                        result = firstNum ** secondNum
                elif currentOperator == "%":
                        if secondNum == 0:
                                return "Error: Cannot Divide By Zero."
                        result = firstNum + (math.ceil((-firstNum / secondNum).real) + complex(0, 1) * math.ceil((-firstNum / secondNum).imag)) * secondNum
                elif currentOperator == "/":
                        if secondNum == 0:
                                return "Error: Cannot Divide By Zero."
                        result = firstNum / secondNum
                elif currentOperator == "*":
                        result = firstNum * secondNum
                elif currentOperator == "-":
                        result = firstNum - secondNum
                elif currentOperator == "+":
                        result = firstNum + secondNum
                        
                if currentOperator == "!":
                        if firstNum.imag == 0 and firstNum.real < 0 and firstNum.real == int(firstNum.real):
                                return "Error: The factorial function is not defined on the negative integers."
                        result = funcEval.gamma(firstNum + 1)
                        expression = expression[:previousIndex + 1] + ('{0:.' + str(precision + 5) + 'f}').format(result) + expression[currentIndex + 1:]
                else:
                        expression = expression[:previousIndex + 1] + complexFormat(result) + expression[nextIndex:]
        return expression
        
def _format(expression):
        expression = expression.replace(" ", "")
        expression = expression.lower()
        expression = expression.replace("×", "*")
        expression = expression.replace("÷", "/")
        expression = expression.replace("--", "+")
        expression = expression.replace("+-", "-")
        expression = expression.replace(")(", ")*(")
        expression = expression.replace("mod", "%")

        # Substitutes in values for mathematical constants
        for index in xrange(0, len(expression)):
                if expression[index] == "e":
                        if index == 0 and expression[index + 1] in _operatorSet:
                                expression = expression.replace("e", ('{0:.' + str(precision + 5) + 'f}').format(math.e))
                        elif index == len(expression) - 1 and expression[index - 1] in _operatorSet:
                                expression = expression.replace("e", ('{0:.' + str(precision + 5) + 'f}').format(math.e))
                        elif expression[index + 1] in {"+", "-", "*", "/", "%", "^", "!"} and expression[index - 1] in _operatorSet:
                                expression = expression.replace("e", ('{0:.' + str(precision + 5) + 'f}').format(math.e))
                        index += 17
        expression = expression.replace("π", ('{0:.' + str(precision + 5) + 'f}').format(math.pi))
        expression = expression.replace("pi", ('{0:.' + str(precision + 5) + 'f}').format(math.pi))

        expression = noScientificNotation(expression)
        
        # Turns "number(expression)" into "number * (expression)"
        symbols = openBrackets.union({"pi", "π"})
        for digit in _digits:
                for symbol in symbols:
                        expression = expression.replace(digit + symbol, digit + "*" + symbol)

        # Parses mathematical functions
        expression = expression.replace("sine", "sin").replace("cosine", "cos").replace("tangent", "tan")
        expression = expression.replace("cosecant", "csc").replace("secant", "sec").replace("cotangent", "cot")
        expression = expression.replace("arc", "a").replace("sqrt", "sqr")
        expression = funcEval.evaluateFunctions(expression)
        return expression

# Public Methods

def noScientificNotation(expression):
        while expression.find("e") > 0 and expression[expression.find("e") - 1] in _digits:
                        index = expression.find("e")
                        # Separate out coefficient and exponent from scientific notation
                        previousOperatorIndex = _indexOfPreviousOperator(expression, index)
                        nextOperatorIndex = _indexOfOperator(expression[index + 1:])
                        if nextOperatorIndex == -1:
                                nextOperatorIndex = _indexOfSpecificCharacters(expression[index:], closeBrackets)
                                if nextOperatorIndex == -1:
                                        nextOperatorIndex = len(expression)
                                else:
                                        nextOperatorIndex += index 
                        else:
                                nextOperatorIndex += index + 1                                
                        coefficient = float(expression[previousOperatorIndex + 1: index])
                        exponent = float(expression[index + 1 : nextOperatorIndex])

                        # Creates new number as string and substitutes it into the expression, readjusting the index for the iterator
                        newNumber = ('{0:.' + str(precision + 5) + 'f}').format(coefficient * (10 ** exponent))
                        expression = expression[:previousOperatorIndex + 1] + newNumber + expression[nextOperatorIndex:]
                        index = previousOperatorIndex + len(newNumber) + 1
        return expression
        
def complexFormat(complexNum):
        if complexNum.imag == 0:
                return '{0:.15f}'.format(complexNum.real)
        elif complexNum.real == 0:
                return '{0:.15f}'.format(complexNum.imag) + "j"
        else:
                num = "<" + '{0:.15f}'.format(complexNum.real) + "+" + '{0:.15f}'.format(complexNum.imag) + "j>"
                return num.replace("+-", "-")

def numberFormat(num):
        if num.find("Error") != -1:
                return num[num.find("Error") : num[num.find("Error"):].find(".") + num.find("Error")]
        
        # Rounds any decimals, removing trailing zeros and representing integers without decimal points
        num = complex(num.replace("i", "j"))
        if num.imag == 0:
                num = ('{0:.' + str(precision) + 'f}').format(num.real)
                # Remove trailing zeroes
                while num[len(num) - 2] != "." and float(num) == float(num[:-1]):
                        num = num[:-1]
                if float(num) == int(float(num)):
                        num = str(int(float(num)))
                return num
        elif num.real == 0:
                num = ('{0:.' + str(precision) + 'f}').format(num.imag)
                # Remove trailing zeroes
                while num[len(num) - 2] != "." and float(num) == float(num[:-1]):
                        num = num[:-1]
                if float(num) == int(float(num)):
                        num = str(int(float(num)))
                return (num + "i").replace("+1i", "+i").replace("-1i", "-i")
        else:
                real = ('{0:.' + str(precision) + 'f}').format(num.real)
                imag = ('{0:.' + str(precision) + 'f}').format(num.imag)
                # Remove trailing zeroes
                while real[len(real) - 2] != "." and float(real) == float(real[:-1]):
                        real = real[:-1]
                while imag[len(imag) - 2] != "." and float(imag) == float(imag[:-1]):
                        imag = imag[:-1]
                if float(real) == int(float(real)):
                        real = str(int(float(real)))
                if float(imag) == int(float(imag)):
                        imag = str(int(float(imag)))
                num = real + "+" + imag + "i"
                num = num.replace("+-", "-")
                return num.replace("+1i", "+i").replace("-1i", "-i")
                
def eval(expression):
        expression = _format(expression)

        # Parenthesis and Brackets
        while _indexOfSpecificCharacters(expression, closeBrackets) != -1:
                # Finds first close bracket symbol, then finds the preceeding open bracket symbol
                closeBracketIndex = _indexOfSpecificCharacters(expression, closeBrackets)
                openBracketIndex = _indexOfPreviousCharacters(expression, {"(", "{", "["}, closeBracketIndex)
                if openBracketIndex < -1:
                        openBracketIndex = -1
                # Evaluates the expression in the brackets and substitutes it into the overall expression
                answer = eval(expression[openBracketIndex + 1 : closeBracketIndex])

                # Accounts for unary negation outside brackets
                if openBracketIndex > 0 and expression[openBracketIndex - 1] == "-":
                        if openBracketIndex > 1 and expression[openBracketIndex - 2] in _operatorSet:
                                answer = eval("-1*(" + answer + ")")
                                openBracketIndex -= 1
                                
                if complex(answer.replace("i", "j")).imag != 0 or complex(answer.replace("i", "j")).real < 0:
                        answer = "<" + answer + ">"
                expression = expression[:openBracketIndex] + answer + expression[closeBracketIndex + 1:]

        # Applies the rest of the order of operations to the expression
        expression = _pemdas(expression, {"!"})
        expression = _pemdas(expression, {"^"})
        expression = _pemdas(expression, {"*", "/", "%"})
        expression = _pemdas(expression, {"+", "-"})
        return numberFormat(expression)
        
def evaluate(expression):
        try:
                return eval(expression)
        except:
                return "Error"




from distutils.core import setup

setup(
    name='AMP',
    version='1.1.3',
    author='Ini Oguntola',
    author_email='ioguntol@gmail.com',
    packages=['AMP', 'AMP.test'],
    scripts=[],
    url='http://github.com/ioguntol/Automatic-Mathematical-Parser/',
    license='LICENSE.txt',
    description='Automatic Mathematical Parser.',
    long_description=open('README.txt').read(),
    install_requires=[
        "Django >= 1.1.1",
        "caldav == 0.1.4",
    ],
)
